path = require 'path'
fs = require 'fs'
os = require 'os'
url = require 'url'
crypto = require 'crypto'
zlib = require 'zlib'

_ = require 'lodash'
request = require 'request'
colors = require 'colors'
semver = require 'semver'
fstream = require 'fstream'
tar = require 'tar'
wrench = require 'wrench'

config = require('../config.coffee').data
util = require 'util'

module.exports = publish = (program) ->
    program
        .command 'publish'
        .description 'Publish a package to the CEX registry'
        .action publishComponent


publishComponent = (options) ->
    cwd = process.cwd()
    packageConfigPath = path.join cwd, config.jsonFileName

    if !config._auth or !config.email
        console.log 'You need to be logged in in order to publish a component'
        return

    if !fs.existsSync packageConfigPath
        console.log "No #{config.jsonFileName.cyan} found in your project"
        return

    try
        data = JSON.parse(fs.readFileSync(packageConfigPath))
        packageData = computePackageData(data)
    catch e
        console.log e.message
        return

    compressPackage cwd, packageData, (stat, tarbuffer) ->
        packageUrl = url.resolve config.paths.registryUrl, packageData.name
        shasum = crypto.createHash("sha1").update(tarbuffer).digest("hex")
        messageBody =
            _id: packageData.name
            name: packageData.name
            description: packageData.description
            'dist-tags': {}
            versions: {}
            readme: data.readme || ""
            maintainers: [{
                name: config._auth.user
                email: config.email
            }]

        messageBody.versions[packageData.version] = packageData
        packageData.maintainers = _.cloneDeep messageBody.maintainers
        tag = packageData.tag || 'latest'
        messageBody['dist-tags'][tag] = packageData.version

        tarballName = "#{packageData.name}-#{packageData.version}.tgz"

        packageData._shasum = shasum

        packageData.dist =
            shasum: shasum
            tarball: "#{packageUrl}/-/#{tarballName}"
                .replace(/^https:\/\//, "http://")

        messageBody._attachments = {}
        messageBody._attachments[tarballName] =
            content_type: 'application/octet-stream'
            data: tarbuffer.toString 'base64'
            length: stat.size

        req = request.put packageUrl, {
            auth: config._auth
        }, (err, response, body) ->
            data = JSON.parse body

            if err or data.error
                console.log 'There has been an error when publishing your
                    package, please try again'
                return

            console.log "Package #{packageData._id.cyan.bold} published
                succesfully!"

        req.write JSON.stringify(messageBody)

computePackageData = (data, cwd) ->
    version = semver.clean data.version
    if !data.name.trim()
        throw new Error "You must specify a name for your package in
            #{config.jsonFileName}"
    if !version
        throw new Error "#{data.version} does not comply with the Semver
            standard"

    data._id = "#{data.name}@#{version}"
    data._from = '.'
    data._npmUser =
        name: config._auth.user
        email: config.email

    data.readme = try
        fs.readFileSync path.join(cwd, 'README.md')
    catch
        "No README"
    data.readmeFilename = 'README.md'

    return data

fixDirs = (entry) ->
    if entry.props.type is 'Directory'
        entry.props.mode |= (entry.props.mode >>> 2) & 0o111
    return true

compressPackage = (packagePath, pkg, cb) ->
    tmpPath = path.join os.tmpDir(), (pkg.name + '-' + pkg.version)
    tmpPackagePath = path.join(tmpPath, 'package')
    tmpArchivePath = path.join tmpPath, 'package.tgz'

    wrench.mkdirSyncRecursive tmpPath, 0o777
    wrench.copyDirSyncRecursive packagePath, tmpPackagePath
    output = fs.createWriteStream tmpArchivePath

    output.on 'finish', ->
        stat = fs.statSync tmpArchivePath
        archive = fs.readFileSync tmpArchivePath

        cb(stat, archive)
    fstream.Reader({path: tmpPackagePath, type: 'Directory', filter: fixDirs})
        .pipe tar.Pack noProprietary: true
        .pipe zlib.createGzip()
        .pipe output
